 /*
    9-26-10
	
	RTC, Pressure, GPS, Accel, Mag
	
	Fault: 1 gps waiting, 2 gps not connected
*/

#include <stdlib.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "types.h"
#include "defs.h"
#include "math.h"	// To calculate altitude
#include "i2c.h"

#define FOSC 8000000
#define BAUD 4800
#define STATUS 2
#define MUXA 6 //S0
#define MUXB 7  //S1

#define BMPCAL	45 // (meters) increase to deacrease pressure
#define TEMPCAL	2775 // (degrees F) increase to increast temp
#define BMP085_R 0xEF
#define BMP085_W 0xEE
#define OSS 0	// Oversampling Setting (note: code is not set up to use other OSS values)

#define sbi(var, mask)   ((var) |= (uint8_t)(1 << mask))
#define cbi(var, mask)   ((var) &= (uint8_t)~(1 << mask))

#define WRITE_sda() DDRC = DDRC | 0b00010000 //SDA must be output when writing
#define READ_sda()  DDRC = DDRC & 0b11101111 //SDA must be input when reading

///============Initialize Prototypes=====//////////////////
void init(void);
void UART_Init(unsigned int ubrr);
void put_char(unsigned char byte);
static int uart_putchar(char c, FILE *stream);
uint8_t uart_getchar(void);
static FILE mystdout = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE);
void delay_ms(uint16_t x);

/////=========Global Variables======////////////////////
volatile short ac1;
volatile short ac2; 
volatile short ac3; 
volatile unsigned short ac4;
volatile unsigned short ac5;
volatile unsigned short ac6;
volatile short b1; 
volatile short b2;
volatile short mb;
volatile short mc;
volatile short md;

volatile unsigned long latitude = 0;
volatile unsigned long longitude = 0;
volatile unsigned long longitude2 = 0;
volatile unsigned long latitude2 = 0;
volatile uint8_t lock = 0;
volatile unsigned long altitude = 0;
volatile long count = 0; 
volatile unsigned long time = 0;
volatile uint8_t sats = 0;
volatile uint8_t fault_count = 0;
volatile long geiger_count = 0;
volatile long temperature = 0;
volatile long temp_t = 0;

///============Function Prototypes=========/////////////////
long BMP085(void);
void BMP085_Calibration(void);
void GPS_enable(void);
void Radio_enable(void);
void openlog_enable(void);
void extra_enable(void);
void parse(void);
void print_GPS(void);
void fault(void);
void radio_print(void);
void radio_print_nolock(void);
void openlog_print(void);

///=========================//////////////////
long bmp085ReadShort(unsigned char address);
long bmp085ReadTemp(void);
long bmp085ReadPressure(void);
void bmp085Convert(long * temperature, long * pressure);

void ADXLInit(void);
int16_t x_accel(void);

uint16_t magnetometer_x(void);

/////===================================////////////////////
ISR(PCINT0_vect)
{
	//Geiger counter trigger	
	//printf("test\n\r");
}

ISR(TIMER1_OVF_vect)
{
	TCNT1 = 34000;	//(256/8MHz)*(65536bits-34000)~=1.009s
	count++;
	fault_count++;
}

ISR(USART_RX_vect) 
{
	fault_count = 0;
	cli();
	
	if(uart_getchar() == '$') 
	{
		if(uart_getchar() == 'G') 
		{
			if(uart_getchar() == 'P')
			{
				if(uart_getchar() == 'G') //start of GPG
				{
					parse();
					
					if(lock == 0x31 || lock == 0x32) //if lock digit equals 1 or 2
					{					
						cbi(PORTB, STATUS);
						delay_ms(50);
						
						//put functions here, decrease delay
						
						//enable and print to radio
						Radio_enable();
						radio_print();
						
						//enable and print to openlog
						openlog_enable();
						radio_print();
						
						GPS_enable();
					} 
					else 
					{
						sbi(PORTB, STATUS);
						Radio_enable();
						radio_print_nolock();
						openlog_enable();
						radio_print_nolock();
						GPS_enable();
					}
				}
			}
		}
	}
	GPS_enable();
	sei();
}

int main(void)
{
	init();
	i2cInit();
	GPS_enable();
	delay_ms(500);
	
	while(1)
	{
		sbi(PORTB, STATUS);
		if(fault_count > 10) fault(); //can't find 11 chars
	}
	return 0;
}

void openlog_print(void)
{
	print_GPS();
	printf("#P%5ld,", BMP085());
	printf("T%4ld,", temp_t);
	printf("m%4d,", magnetometer_x());
	printf("a%4d,", x_accel());
	printf("%ld@\r\n", count);
}

void radio_print(void)
{
	printf("$%2d,", sats);
	printf("%6ld,", time);
	printf("N%ld.", latitude2);
	printf("%ld,", latitude);
	printf("W%ld.", longitude2);
	printf("%ld,", longitude);
	printf("A%5ld,", altitude);
	printf("P%5ld,", BMP085());
	printf("T%4ld,", temp_t);
	printf("m%4d,", magnetometer_x());
	printf("a%4d,", x_accel());
	printf("%ld@", count);
	printf("\r\n");
}

void radio_print_nolock(void)
{
	printf("$P%5ld,", BMP085());
	printf("T%4ld,", temp_t);
	printf("mag%4d,", magnetometer_x());
	printf("acc%4d,", x_accel());
	printf("%ld@\r\n", count);
}

void fault(void)
{
	Radio_enable();
	extra_enable();
	printf("$fault#P%5ld,", BMP085());
	printf("T%4ld,", temp_t);
	printf("m%4d,", magnetometer_x());
	printf("a%4d,", x_accel());
	printf("%ld@\r\n", count);
	
	openlog_enable();
	printf("$fault#P%5ld,", BMP085());
	printf("T%4ld,", temp_t);
	printf("m%4d,", magnetometer_x());
	printf("a%4d,", x_accel());
	printf("%ld@\r\n", count);
	GPS_enable();
}

void geiger(void)
{
	//while(!(PINB & (1<<PINB1)))gegier_count++;
}

long BMP085(void)
{
	long temperature = 0;
	//long temp_t = 0;
	long pressure = 0;
	long altitude = 0;
	double temp = 0;

	bmp085Convert(&temperature, &pressure);
	
	temp_t = temperature + TEMPCAL;

	//printf("T%4ld,", temp_t);
	
	temp = (double) pressure/101325;
	temp = 1-pow(temp, 0.19029);
	altitude = (round(10458*temp))-BMPCAL;
	
	return altitude;
}

void BMP085_Calibration(void)
{
	ac1 = bmp085ReadShort(0xAA);
	ac2 = bmp085ReadShort(0xAC);
	ac3 = bmp085ReadShort(0xAE);
	ac4 = bmp085ReadShort(0xB0);
	ac5 = bmp085ReadShort(0xB2);
	ac6 = bmp085ReadShort(0xB4);
	b1 = bmp085ReadShort(0xB6);
	b2 = bmp085ReadShort(0xB8);
	mb = bmp085ReadShort(0xBA);
	mc = bmp085ReadShort(0xBC);
	md = bmp085ReadShort(0xBE);
	
	/*printf("\tAC1 = %d\n", ac1);
	printf("\tAC2 = %d\n", ac2);
	printf("\tAC3 = %d\n", ac3);
	printf("\tAC4 = %d\n", ac4);
	printf("\tAC5 = %d\n", ac5);
	printf("\tAC6 = %d\n", ac6);
	printf("\tB1 = %d\n", b1);
	printf("\tB2 = %d\n", b2);
	printf("\tMB = %d\n", mb);
	printf("\tMC = %d\n", mc);
	printf("\tMD = %d\n", md);
	printf("------------------------\n\n");*/
	
	delay_ms(900);
}

// bmp085ReadShort will read two sequential 8-bit registers, and return a 16-bit value
// the MSB register is read first
// Input: First register to read
// Output: 16-bit value of (first register value << 8) | (sequential register value)
long bmp085ReadShort(unsigned char address)
{
	char msb, lsb;
	long data;
	
	i2cSendStart();
	i2cWaitForComplete();
	
	i2cSendByte(BMP085_W);	// write 0xEE
	i2cWaitForComplete();
	
	i2cSendByte(address);	// write register address
	i2cWaitForComplete();
	
	i2cSendStop();
	i2cSendStart();
	
	i2cSendByte(BMP085_R);	// write 0xEF
	i2cWaitForComplete();
	
	i2cReceiveByte(TRUE);
	i2cWaitForComplete();
	msb = i2cGetReceivedByte();	// Get MSB result
	i2cWaitForComplete();
	
	i2cReceiveByte(FALSE);
	i2cWaitForComplete();
	lsb = i2cGetReceivedByte();	// Get LSB result
	i2cWaitForComplete();
	
	i2cSendStop();
	
	data = msb << 8;
	data |= lsb;
	
	return data;
}

long bmp085ReadTemp(void)
{
	i2cSendStart();
	i2cWaitForComplete();
	
	i2cSendByte(BMP085_W);	// write 0xEE
	i2cWaitForComplete();
	
	i2cSendByte(0xF4);	// write register address
	i2cWaitForComplete();
	
	i2cSendByte(0x2E);	// write register data for temp
	i2cWaitForComplete();
	
	i2cSendStop();
	
	delay_ms(40);	// max time is 4.5ms
	
	return (long) bmp085ReadShort(0xF6);
}

long bmp085ReadPressure(void)
{
	long pressure = 0;
	
	i2cSendStart();
	i2cWaitForComplete();
	
	i2cSendByte(BMP085_W);	// write 0xEE
	i2cWaitForComplete();
	
	i2cSendByte(0xF4);	// write register address
	i2cWaitForComplete();
	
	i2cSendByte(0x34);	// write register data for temp
	i2cWaitForComplete();
	
	i2cSendStop();
	
	delay_ms(15);	// max time is 4.5ms
	
	pressure = bmp085ReadShort(0xF6);
	pressure &= 0x0000FFFF;
	
	return pressure;
}

void bmp085Convert(long* temperature, long* pressure)
{
	long ut;
	long up;
	long z1, z2, x1, x2, b5, b6, x3, b3, p;
	unsigned long b4, b7;
	
	ut = bmp085ReadTemp();
	ut = bmp085ReadTemp();	// some bug here, have to read twice to get good data
	up = bmp085ReadPressure();
	up = bmp085ReadPressure();
	
	z1 = ((long)ut - ac6) * ac5 >> 15;
	z2 = ((long) mc << 11) / (z1 + md);
	b5 = z1 + z2;
	*temperature = (b5 + 8) >> 4;
	
	//printf("T%4ld,", ut);
	
	b6 = b5 - 4000;
	x1 = (b2 * (b6 * b6 >> 12)) >> 11;
	x2 = ac2 * b6 >> 11;
	x3 = x1 + x2;
	b3 = (((int32_t) ac1 * 4 + x3) + 2)/4;
	x1 = ac3 * b6 >> 13;
	x2 = (b1 * (b6 * b6 >> 12)) >> 16;
	x3 = ((x1 + x2) + 2) >> 2;
	b4 = (ac4 * (unsigned long) (x3 + 32768)) >> 15;
	b7 = ((unsigned long) up - b3) * (50000 >> OSS);
	p = b7 < 0x80000000 ? (b7 * 2) / b4 : (b7 / b4) * 2;
	x1 = (p >> 8) * (p >> 8);
	x1 = (x1 * 3038) >> 16;
	x2 = (-7357 * p) >> 16;
	*pressure = p + ((x1 + x2 + 3791) >> 4);
}

void print_GPS(void)
{
	printf("$%d,", sats);
	printf("%ld,", time);
	printf("N%ld.", latitude2);
	printf("%ld,", latitude);
	printf("W%ld.", longitude2);
	printf("%ld,", longitude);
	printf("A%ld&", altitude);
}

void parse(void)
{
	for(int i=0; i<3; i++)
	{
		uart_getchar(); //GA,
	}
	
	//grab the UTC 6 digits
	long k0 = uart_getchar();
	long k1 = uart_getchar(); 
	long k2 = uart_getchar();
	long k3 = uart_getchar();
	long k4 = uart_getchar();
	long k5 = uart_getchar();
	
	for(int m=0; m<4; m++)
	{
		uart_getchar();
	}
	
	//grab the N first 4 integer digits
	long m0 = uart_getchar();
	long m1 = uart_getchar(); 
	long m2 = uart_getchar();
	long m3 = uart_getchar();
	
	uart_getchar(); //decimal
	
	//grab the N 5 decimal digits
	long m0a = uart_getchar();
	long m1a = uart_getchar(); 
	long m2a = uart_getchar();
	long m3a = uart_getchar();
	long m4a = uart_getchar();
	
	for(int j=0; j<3; j++)
	{
		uart_getchar();
	}
	
	//grab the W first 5 integer digits
	long mw0 = uart_getchar();
	long mw1 = uart_getchar(); 
	long mw2 = uart_getchar();
	long mw3 = uart_getchar();
	long mw4 = uart_getchar();
	
	uart_getchar(); //decimal
	
	//grab W 5 digits
	long m0b = uart_getchar();
	long m1b = uart_getchar(); 
	long m2b = uart_getchar();
	long m3b = uart_getchar();
	long m4b = uart_getchar();
	
	uart_getchar(); //comma
	uart_getchar(); //W
	uart_getchar(); //comma

	//grab lock 
	lock = uart_getchar();
	uart_getchar();
	uint8_t sat1 = uart_getchar();
	uint8_t sat2 = uart_getchar();
	
	for(int l=0; l<6; l++)
	{
		uart_getchar();
	}
	
	//Grab altitude
	long m0c = uart_getchar();
	long m1c = uart_getchar(); 
	long m2c = uart_getchar();
	long m3c = uart_getchar();
	long m4c = uart_getchar();
	
	//////////////////////////////////////////////////
	/////////Calculate////////////////////////////////
	//////////////////////////////////////////////////
	
	// UTC digits
	time = (k0-48)*(100000);
	time = (k1-48)*(10000)+time;
	time = (k2-48)*(1000)+time;
	time = (k3-48)*(100)+time;
	time = (k4-48)*(10)+time;
	time = (k5-48)+time;
	
	//N first for integer digits
	latitude2 = (m0-48)*(1000);
	latitude2 = (m1-48)*(100)+latitude2;
	latitude2 = (m2-48)*(10)+latitude2;
	latitude2 = (m3-48)+latitude2;
	
	//N last decimal digits
	latitude = (m0a-48)*(10000);
	latitude = (m1a-48)*(1000)+latitude;
	latitude = (m2a-48)*(100)+latitude;
	latitude = (m3a-48)*(10)+latitude;
	latitude = (m4a-48)+latitude;
	
	//W first five integer digits
	longitude2 = (mw0-48)*(10000);
	longitude2 = (mw1-48)*(1000)+longitude2;
	longitude2 = (mw2-48)*(100)+longitude2;
	longitude2 = (mw3-48)*(10)+longitude2;
	longitude2 = (mw4-48)+longitude2;
	
	//lock
	//lock = bbb;
	
	//satellites
	sats = (sat1-48)*(10);
	sats = (sat2-48)+sats;
	
	//W last decimal digits 
	longitude = (m0b-48)*(10000);
	longitude = (m1b-48)*(1000)+longitude;
	longitude = (m2b-48)*(100)+longitude;
	longitude = (m3b-48)*(10)+longitude;
	longitude = (m4b-48)+longitude;
	
	altitude = (m0c-48)*(10000);
	altitude = (m1c-48)*(1000)+altitude;
	altitude = (m2c-48)*(100)+altitude;
	altitude = (m3c-48)*(10)+altitude;
	altitude = (m4c-48)+altitude;
	
}

uint16_t magnetometer_x(void)
{	
	uint8_t xh, xl, yh, yl, zh, zl;
	
	i2cSendStart();
	i2cWaitForComplete();
	i2cSendByte(0x3C);    //write to HMC
	i2cWaitForComplete();
	i2cSendByte(0x02);    //mode register
	i2cWaitForComplete();
	i2cSendByte(0x00);    //continuous measurement mode
	i2cWaitForComplete();
	i2cSendStop();
	
	//must read all six registers plus one to move the pointer back to 0x03
	i2cSendStart();
	i2cWaitForComplete();
	i2cSendByte(0x3D);          //read from HMC
	i2cWaitForComplete();
	i2cReceiveByte(TRUE);
	i2cWaitForComplete();
	xh = i2cGetReceivedByte();	//x high byte
	i2cWaitForComplete();
	
	i2cReceiveByte(TRUE);
	i2cWaitForComplete();
	xl = i2cGetReceivedByte();	//x low byte
	i2cWaitForComplete();
	uint16_t x_mag = xl|(xh << 8);
	
	i2cReceiveByte(TRUE);
	i2cWaitForComplete();
	yh = i2cGetReceivedByte();	//y high byte
	i2cWaitForComplete();
	
	i2cReceiveByte(TRUE);
	i2cWaitForComplete();
	yl = i2cGetReceivedByte();	//y low byte
	i2cWaitForComplete();
	
	i2cReceiveByte(TRUE);
	i2cWaitForComplete();
	zh = i2cGetReceivedByte();	
	i2cWaitForComplete();      //z high byte
	
	i2cReceiveByte(TRUE);
	i2cWaitForComplete();
	zl = i2cGetReceivedByte();	//z low byte
	i2cWaitForComplete();
	
	i2cSendByte(0x3D);         //must reach 0x09 to go back to 0x03
	i2cWaitForComplete();
	i2cSendStop();
	
	return x_mag;	
}

void ADXLInit(void)
{
	
	//initialize
	i2cSendStart();
	i2cWaitForComplete();
	i2cSendByte(0xA6);    //write to ADXL
	i2cWaitForComplete();
	i2cSendByte(0x2D);    //power register
	i2cWaitForComplete();
	i2cSendByte(0x08);    //measurement mode
	i2cWaitForComplete();
	i2cSendStop();
	
	i2cSendStart();
	i2cWaitForComplete();
	i2cSendByte(0xA6);    //write to ADXL
	i2cWaitForComplete();
	i2cSendByte(0x31);    //data format
	i2cWaitForComplete();
	i2cSendByte(0x0B);    //full resolution and 16g
	i2cWaitForComplete();
	i2cSendStop();
}
		
int16_t x_accel(void)
{		
	ADXLInit();	
		
	char lsb, msb;

	i2cSendStart();
	i2cWaitForComplete();
	
	i2cSendByte(0xA6);	// write to this I2C address, R/*W cleared
	i2cWaitForComplete();
	
	i2cSendByte(0x32);	//x axis, 0x34 y, 0x36 z
	i2cWaitForComplete();
	
	i2cSendStart();
	
	i2cSendByte(0xA7); // read from this I2C address, R/*W Set
	i2cWaitForComplete();
	
	i2cReceiveByte(TRUE);
	i2cWaitForComplete();
	lsb = i2cGetReceivedByte(); //Read the LSB data
	i2cWaitForComplete();

	i2cReceiveByte(FALSE);
	i2cWaitForComplete();
	msb = i2cGetReceivedByte(); //Read the MSB data
	i2cWaitForComplete();
	
	i2cSendStop();
	
	return( (msb<<8) | lsb);
}

/*********************
 ****MUX enables****
 *********************/
 
void GPS_enable(void)
{
	sbi(PORTD, MUXA);
	cbi(PORTD, MUXB);
	//delay_ms(10);
}

void Radio_enable(void)
{	
	cbi(PORTD, MUXA);
	sbi(PORTD, MUXB);
	delay_ms(10);
}

void openlog_enable(void)
{	
	cbi(PORTD, MUXA);
	cbi(PORTD, MUXB);
	delay_ms(10);
}

void extra_enable(void)
{	
	sbi(PORTD, MUXA);
	sbi(PORTD, MUXB);
	delay_ms(10);
}

/*********************
 ****Initialize****
 *********************/
 
void init (void)
{
    //1 = output, 0 = input
    DDRB = 0b00101100; //outs on SCK, MOSI, LED (PB2)
    DDRC = 0b00010101; //outs on MODE, RTB
    DDRD = 0b11011010; //outs on TX, 
	sbi(PORTD, 3); //keep BMP from reseting
	//PORTC = 0b00110000; //pullups on the I2C bus
	
	//LED off
	cbi(PORTB, STATUS);
	
	//UART_Init((unsigned int)(FOSC/16/(BAUD*2)-1));
	// ocillator fq/16/baud rate -1	
	UART_Init(103);//4800, 51=9600
	
	// SETTING USART RX ISR
	UCSR0B |= (1 << RXCIE0); 
	
	delay_ms(1000);

	BMP085_Calibration();

	//configure Geiger
	PCICR = 0x01;
	PCMSK0 = 0x02; //enable PCINT1
		
	//configure GPS
	//GPS_enable();
	//delay_ms(10);
	//printf("$PSRF103,05,00,00,01*21\n\r");


	//configure I2C devices
	//ADXLInit();
	//delay_ms(5);
	//BMP085_Calibration();
	
	// SETTING TIMER 1:
	// normal mode
	TCCR1A = 0x00;
	// Set to clk/256 => (256/8MHz)*(65536bits)~=2.09s
	TCCR1B |= (1<<CS12);
	//enable overflow interrupt
	TIMSK1 |= (1<<TOIE1);
	TCNT1 = 34000;	//(256/8MHz)*(65536bits-34000)~=1.009s
	
	//enable global interrupts
	sei();
	
}


void UART_Init( unsigned int ubrr)
{
	// Set baud rate 
	UBRR0H = ubrr>>8;
	UBRR0L = ubrr;
	
	// Enable receiver and transmitter 
	UCSR0A = (0<<U2X0);
	UCSR0B = (1<<RXEN0)|(1<<TXEN0);
	
	// Set frame format: 8 bit, no parity, 1 stop bit,   
	UCSR0C = (1<<UCSZ00)|(1<<UCSZ01);
	
	stdout = &mystdout; //Required for printf init
}

static int uart_putchar(char c, FILE *stream)
{
    if (c == '\n') uart_putchar('\r', stream);
  
    loop_until_bit_is_set(UCSR0A, UDRE0);
    UDR0 = c;
    
    return 0;
}

uint8_t uart_getchar(void)
{
    while( !(UCSR0A & (1<<RXC0)) );
    return(UDR0);
}

void put_char(unsigned char byte)
{
	/* Wait for empty transmit buffer */
	while ( !( UCSR0A & (1<<UDRE0)) );
	/* Put data into buffer, sends the data */
	UDR0 = byte;
}
